#include <stdio.h>
#include <stdlib.h>	// atof
#include <math.h>	// cbrt
#include <string.h>	// strcmp
#define G0 9.80665
#define ATM 101300	// Pa

struct tank_struct {
	double length; double radius; double diameter;	// internal volume
	double thickness; double inner_volume; double empty_mass;
};

void handle_command_line_parameters(int argc, char ** argv);
void display_tank(char *prop_name, struct tank_struct *tank);
void display_parameters();
void display_usage(char *argv0);
void mass_ratio_from_Isp_and_DV();
double stoichiometric_ratio(int fuel, int oxidizer);
double propellant_mass_to_volume(int propellant, double mass);
enum propellant { LOX, RP1, E85, H2, ETHANOL925 };

/* USING SI UNITS *
 * mass: kg
 * weight: N
 * thrust: N
 * length: m
 * volume: m^3
 * pressure: Pa
 * g0 = 9.8 m/s
 */

/* depending on mission (Delta-V) and engine parameters (chosen fuel, chamber
 * pressure, engine nozzle, giving exhaust velocity), the mass ratio of the
 * rocket is given by the rocket equation:
 * e^(-delta_v/exhaust_velocity) is dry mass / wet mass
 *
 * the exhaust velocity can also be obtained from Isp, the relation between them
 * being: Isp * G0 = exhaust velocity. G0 is the standard gravity constant.
 *
 * so the mass ratio from the Isp is:
 * 1 / e^(-Delta V / (Isp * G0))	with delta V in m/s instead of km/s
 */

double deltaV = 8000;		// km/s
double isp = 340;		// s
// 8.855 %

/***************************** INPUT DATA *****************************/

/* ratios */
double mass_ratio = 11.3;		// wet mass / dry mass
// WARNING: mass ratio is recomputed if deltaV and isp are non zero
double engine_thrust_to_weight = 100.0;
double overall_thrust_to_weight = 1.25;	// ignition thrust ratio
double length_to_diameter = 10.0;	// size of the propellant tanks
					// ltd = (h1 + h2) / d, with d = 2r

double payload_mass = 0.0199;		// < 20g
double avionics_board_mass = 0.06;	// the processor board for avionics
double avionics_battery_mass = 0.15;	// bat. required for all but vector actuator
double sensors_mass = 0.05;		// 50g should be more than enough
/* accelerometer or gyroscope = 50mg
 * with breakout: 1.4g
 * pressure sensor: 4g (without tubing and pitot)
 * camera: 2g?
 */

double first_dry_mass_estimation = 10.;	// first iteration

int number_of_iterations = 800;
int quiet = 0;
int making_graph = 0;

/* fuel data: it is also required to change the mass ratio or isp above when
 * changing propellants below. Pressure is used to calculate roughly the tank's
 * wall's thickness 1mm per 10 standard atmospheres of pressure. Tank insulation
 * is PU foam, 22mW/m.K, 24kg/m^3 */
int fuel = E85;		char *fuel_name = "E85";
double fuel_pressure = 1 * ATM;
double fuel_total_venting = 1.;			// 1 + percent of vented fuel by mass
double fuel_tank_insulation_thickness = 0;	// 0 for no insulation (meters)
int oxidizer = LOX;	char *oxidizer_name = "LOX";
double oxidizer_pressure = 10 * ATM;
double oxidizer_total_venting = 1.020;		// 1 + percent of vented oxidizer by mass
double oxidizer_tank_insulation_thickness = 0.005;// 0 for no insulation (meters)

/********************** CODE BEGINNING **************************/

double engine_mass(double required_thrust) {
	return required_thrust / engine_thrust_to_weight / G0;
}

/* TODO: iterate on tank diameter, because it's currently the inner diameter
 * that is equal for both tanks, not the outer, which is what we need */
/* tank 1 is fuel, tank 2 is oxidizer */
void tank_data(double prop1_volume, double prop1_pressure, struct tank_struct *tank1,
		double prop2_volume, double prop2_pressure, struct tank_struct *tank2) {
	/* volume of a tank is PI * r^2 * h  +  4 * PI * r^3 / 3
	 * h = h_fuel + h_ox and volume_fuel / volume_ox = h_fuel / h_ox
	 */
	double surface;
	double propellants_volume = prop1_volume + prop2_volume;
	double total_tanks_height;
#ifdef DEBUG
	double tmp;
#endif

	tank1->thickness = 0.0005;// + 0.001 * prop1_pressure / 1013000;
	tank2->thickness = 0.001 + 0.001 * prop2_pressure / 1013000;

	tank1->radius = cbrt(propellants_volume /
			(2. * M_PI * length_to_diameter - 4. * M_PI / 3.));
	tank2->radius = tank1->radius;
	tank1->diameter = tank1->radius * 2.;
	tank2->diameter = tank1->diameter;

	total_tanks_height = length_to_diameter * tank1->diameter;
	tank1->length = total_tanks_height / (1. + 1. / (prop1_volume / prop2_volume));
	tank2->length = total_tanks_height - tank1->length;

	tank1->inner_volume = prop1_volume;
	tank2->inner_volume = prop2_volume;

#ifdef DEBUG
	/* Checking volumes. Approximations make equality fail on result */
	tmp = 4. * M_PI * tank1->radius * tank1->radius * tank1->radius / 3. +
		M_PI * tank1->radius * tank1->radius * (tank1->length - tank1->diameter);
	if (tmp != prop1_volume)
		fprintf(stderr, "miscalculated volume for tank 1: %g instead of %g\n",
				tmp, prop1_volume);
	tmp = 4. * M_PI * tank2->radius * tank2->radius * tank2->radius / 3. +
		M_PI * tank2->radius * tank2->radius * (tank2->length - tank2->diameter);
	if (tmp != prop2_volume)
		fprintf(stderr, "miscalculated volume for tank 2: %g instead of %g\n",
				tmp, prop2_volume);
#endif

	/* now we need the surface of the tank for weight calculation:
	 * surface of sphere is 4 PI * R^2, of cylinder is 2 PI * R * H
	 * tank surface is thus 2 PI * R (2R + H), with H = L - 2R
	 * so it is 2 PI R L = PI D L
	 * aluminium density is 2700 kg/m^3
	 * composite carbon/epoxy density is 1600 kg/m^3
	 * polypropylene density is 855 kg/m^3
	 * HDPE density is 950 kg/m^3
	 */
	surface = M_PI * tank1->diameter * tank1->length;
	//tank1->empty_mass = 2700. * surface * tank1->thickness;
	tank1->empty_mass = 950. * surface * tank1->thickness;
	tank1->empty_mass += fuel_tank_insulation_thickness * surface * 24;
	/* this is an approximation, the real volume is external diameter
	 * cylinder minus internal diameter cylinder volumes */
	
	surface = M_PI * tank2->diameter * tank2->length;
	tank2->empty_mass = 2700. * surface * tank2->thickness;
	tank2->empty_mass += oxidizer_tank_insulation_thickness * surface * 24;
}

/* rocket main structure, aircraft staging mechanism, avionics empty bay,
 * fairing and its mechanism, inter-tank skirt or external skin */
double structure_mass(double vehicle_length, double vehicle_width) {
	double mass = .002 * .015 * vehicle_length * 5 * 2700;	// truss
	mass += .15;	// aircraft staging
	// skirts assimilated to cylinders of 1mm thickness, 10cm spare for engine and bay
	mass += (2. * vehicle_width + 0.1) * M_PI * vehicle_width * 1600 * 0.001;
	// fairing assimilated as a 2mm-thick half-sphere, and mechanism for 100g
	mass += M_PI * vehicle_width * vehicle_width / 2. * 2700 * .002 + .1;
	// NOTE: fairing and skirts could easily be done in composite materials,
	// saving half their weight at least and providing more strength
	return mass;
}

double wiring_mass(double vehicle_length) {
	return vehicle_length * .07;
}

/* mass for the thrust vector actuators (i.e. servo mechanism or hydraulic
 * system) + battery required for the chosen system. */
double actuators_mass(double engine_thrust) {
	return engine_thrust / 10000. +		// servo actuators
		engine_thrust / 15000. ;	// battery
}

void iterate() {
	int i=0;
	double thrust, propellant_mass, fuel_mass, oxidizer_mass;
	double fuel_volume, oxidizer_volume, vehicle_length;
	double m_engine, m_actuators, m_wiring, m_structure;	// loop values
	struct tank_struct tank_fuel, tank_oxidizer;

	double vehicle_dry_mass = first_dry_mass_estimation;	
#ifndef ITERATE_ON_DRY
	double vehicle_wet_mass = vehicle_dry_mass * mass_ratio;
#else
	double vehicle_wet_mass;
#endif
	double stoich_ratio = stoichiometric_ratio(fuel, oxidizer);

	do {
#ifdef ITERATE_ON_DRY
		double next_dry_mass = 0.0;
		vehicle_wet_mass = vehicle_dry_mass * mass_ratio;
#else
		double next_wet_mass;
		vehicle_dry_mass = vehicle_wet_mass / mass_ratio;
#endif
		propellant_mass = vehicle_wet_mass - vehicle_dry_mass;
		thrust = vehicle_wet_mass * G0 * overall_thrust_to_weight;
		/* stoich_ratio = ox mass / fuel mass
		 * prop mass = fuel mass + ox mass
		 * prop mass = fuel mass + fuel mass * stoich ratio
		 * prop mass = fuel mass ( 1 + stoich ratio )
		 * fuel mass = prop mass / ( 1 + stoich ratio )
		 *
		 * prop mass = ox mass / stoich ratio + ox mass
		 * prop mass = ox mass ( 1 + 1 / stoich ratio )
		 * ox mass = prop mass / (1 + 1 / stoich ratio )
		 */
		fuel_mass = propellant_mass / (1. + stoich_ratio);
		fuel_mass *= fuel_total_venting;
		oxidizer_mass = propellant_mass / (1. + 1. / stoich_ratio);
		oxidizer_mass *= oxidizer_total_venting;
		//fprintf(stdout,"fuel mass: %g, oxidizer mass: %g, propellant mass: %g\n",
		//		fuel_mass, oxidizer_mass, propellant_mass);
		fuel_volume = propellant_mass_to_volume(fuel, fuel_mass);
		oxidizer_volume = propellant_mass_to_volume(oxidizer, oxidizer_mass);
		tank_data(fuel_volume, fuel_pressure, &tank_fuel,
				oxidizer_volume, oxidizer_pressure, &tank_oxidizer);
		vehicle_length = tank_fuel.length + tank_oxidizer.length; // roughly

		m_engine = engine_mass(thrust);
		m_actuators = actuators_mass(thrust);
		m_wiring = wiring_mass(vehicle_length);
		m_structure = structure_mass(vehicle_length, tank_fuel.diameter);

#ifdef ITERATE_ON_DRY
		next_dry_mass =
#else
		next_wet_mass = propellant_mass +
#endif
			/* fixed stuff */
			payload_mass + avionics_board_mass + sensors_mass +
			avionics_battery_mass +
			/* variable stuff */
			tank_fuel.empty_mass + tank_oxidizer.empty_mass +
			m_engine + m_actuators + m_wiring + m_structure;

#ifdef ITERATE_ON_DRY
#if DEBUG
		fprintf(stdout, "DRY MASS: %g kg, vehicle size: %g x %g m\n",
				next_dry_mass, vehicle_length, tank_fuel.diameter);
#endif
		vehicle_dry_mass = next_dry_mass;
#else
#if DEBUG
		fprintf(stdout, "DRY MASS: %g kg, tanks size: %g x %g m\n",
				vehicle_dry_mass, vehicle_length, tank_fuel.diameter);
#endif
		vehicle_wet_mass = next_wet_mass;
#endif
		i++;
	} while (i < number_of_iterations) ;
	if (making_graph) {
		/* print out the iterated parameter and result for graph data */
		fprintf(stdout, "%g\t%g\t%g\n",
				overall_thrust_to_weight, 
			       	thrust / (vehicle_dry_mass * G0),
				thrust);
	} else {
		fprintf(stdout, "DRY MASS: %g kg, tanks size: %g x %g m\n",
				vehicle_dry_mass, vehicle_length, tank_fuel.diameter);
		fprintf(stdout, "WET mass: %g kg\n", vehicle_wet_mass);
	}
	if (!quiet) {
		fprintf(stdout, "\nEngine thrust: %g N, engine mass: %g kg\n", thrust, m_engine);
		fprintf(stdout, "Final thrust-to-weight-ratio: %g G\n",
			       	thrust / (vehicle_dry_mass * G0));
		fprintf(stdout, "Actuators mass: %g kg, wiring mass: %g kg\nStructure mass: %g kg\n",
				m_actuators, m_wiring, m_structure);
		display_tank(oxidizer_name, &tank_oxidizer);
		display_tank(fuel_name, &tank_fuel);
	}
}

int main(int argc, char **argv) {
	handle_command_line_parameters(argc, argv);
	mass_ratio_from_Isp_and_DV();	// by default, recompute it
#ifdef MAKE_GRAPH
	making_graph = 1;	quiet = 1;
#endif
	if (!quiet) {
		display_parameters();
#ifdef ITERATE_ON_DRY
		fprintf(stdout, "%d iterations on dry mass\n",
				number_of_iterations);
#else
		fprintf(stdout, "%d iterations on wet mass\n",
				number_of_iterations);
#endif
	}
	
#ifdef MAKE_GRAPH
	/* print a comment for the data output, indicating what data is about */
	fprintf(stdout, "# overall thrust_to_weight\tthrust\tfinal acceleration\n");
	/* iterate over a parameter. Change it here if needed */
	for (overall_thrust_to_weight = 1.; overall_thrust_to_weight < 3.; overall_thrust_to_weight += .05)
#endif
		iterate();
	return 0;
}

void display_tank(char *prop_name, struct tank_struct *tank) {
	fprintf(stdout, "%s tank info:\n", prop_name);
	fprintf(stdout, "\tdiameter: %g m, length: %g m\n\tthickness: %g m, volume: %g m^3, mass: %g kg\n",
			tank->diameter, tank->length, tank->thickness, tank->inner_volume, tank->empty_mass);
}

void display_parameters() {
	fprintf(stdout, "----------- Input parameters -----------\n");
	fprintf(stdout, "Mass ratio: %g. Length-to-diameter ratio (tanks): %g\n",
			mass_ratio, length_to_diameter);
	fprintf(stdout, "Engine thrust-to-weight ratio: %g. Overall thrust-to-weight ratio: %g\n",
			engine_thrust_to_weight, overall_thrust_to_weight);
	fprintf(stdout, "Payload mass: %g kg. Avionics fixed mass: %g kg. Sensors mass: %g kg\n",
			payload_mass, avionics_board_mass, sensors_mass);
	fprintf(stdout, "----------------------------------------\n");
}

/************************* PROPELLANT DATA ****************************/

/* ratio by mass */
double stoichiometric_ratio(int fuel, int oxidizer) {
	switch (oxidizer) {
		case LOX:
			switch (fuel) {
				case RP1:
					return 2.56;	// is that right?
				case E85:
					return 2.26;
				case H2:
					return 8.0;
				case ETHANOL925:
					return 2.19;
			}
	}
	fprintf(stderr, "stoichiometric ratio unknow for selected fuels\n");
	return 0;
}

double propellant_mass_to_volume(int propellant, double mass) {
	switch (propellant) {
		case LOX:
			return mass / 1140.;
		case RP1:
			return mass / 806.;
		case E85:
			return mass / 780.;
		case H2:
			return mass / 71.;
		case ETHANOL925:
			return mass / 800.;
	}
	fprintf(stderr, "unknown propellant used %d\n", propellant);
	return 0;
}

void mass_ratio_from_Isp_and_DV() {
	if (!deltaV || !isp) return;
	mass_ratio = 1. / exp(- deltaV / (isp * G0));
	if (!quiet) {
		fprintf(stdout, "\nUsing computed mass ratio of %g (from Isp=%g and DeltaV=%g)\n",
				mass_ratio, isp, deltaV);
		fprintf(stdout, "Propellant is %g%% by mass.\n\n", (1.-1./mass_ratio)*100.);
	}
}

/************************* COMMAND LINE ****************************/

void handle_command_line_parameters(int argc, char ** argv) {
	int i;
	if (argc <= 1) return;
	if (!strcmp(argv[1], "-h"))
		display_usage(*argv);
	if (!strcmp(argv[1], "-q")) {
		quiet = 1;
		i = 2;
		if (argc == 2) return;
	} else 	i = 1;
	if (argc - i <= 1) display_usage(*argv);
	do {
		double val = atof(argv[i+1]);
		if (val == 0.0) display_usage(*argv);
		switch (argv[i][1]) {
			case '\0':
				break;
			case 'r':
				mass_ratio = val;
				break;
			case 'e':
				engine_thrust_to_weight = val;
				break;
			case 'o':
				overall_thrust_to_weight = val;
				break;
			case 'l':
				length_to_diameter = val;
				break;
			case 'p':
				payload_mass = val;
				break;
			case 'a':
				avionics_board_mass = val;
				break;
			default:
				fprintf(stderr, "Unhandled option %s\n", argv[i]);
				display_usage(*argv);
		}
		i+=2;
	} while (i < argc-1);
}

void display_usage(char *argv0) {
	fprintf(stderr, "Usage: %s [-r mass_ratio] [-e engine_thrust_to_weight]\n"
			"\t[-o overall_thrust_to_weight] [-l length_to_diameter]\n"
			"\t[-p payload_mass] [-a avionics_mass]\n",
			argv0);
	exit(1);
}

